
import com.sun.opengl.util.Animator;
import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureIO;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.nio.IntBuffer;
import java.util.Scanner;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;

/**
 * @author Jeremy Smith
 */
public class CS752_Screen extends GLCanvas implements GLEventListener, MouseListener, MouseMotionListener, KeyListener
{
    private int shaderprogram;
    private final Animator animator = new Animator(this);
    private Texture colorMap, normalMap;
    private CS752_Surface surface;
    
    private float white[] = {1.0f, 1.0f, 1.0f, 1};
    private float gray[] = {0.7f, 0.7f, 0.7f, 1};
    private float darkgray[] = {0.3f, 0.3f, 0.3f, 1};
    private float black[] = {0.0f, 0.0f, 0.0f, 1};
    float lightPosition[] = {-10.0f, 10.0f, 10.0f, 1};
    float shininess[] = {10.0f};
    
    static final int NOTHING = 0, SELECT = 1;
    int cmd = NOTHING;
    int mouseX, mouseY;
    float rotationX = 0.0f, rotationY = 0.0f, rotationZ = 0.0f
        , fixedRotationX = 45.0f, fixedRotationY = 45.0f, fixedRotationZ = 0.0f;
    
    public CS752_Screen() 
    {
        addGLEventListener(this);
        addMouseListener(this);
        addMouseMotionListener(this);
        addKeyListener(this);
        animator.start();
    }

    public void init(GLAutoDrawable drawable) 
    {
        GL gl = drawable.getGL();

        //setup GL properties
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glDepthFunc(GL.GL_LESS);
        gl.glShadeModel(GL.GL_SMOOTH);
        gl.glClearColor(0.7f, 0.7f, 0.7f, 0.0f);
        
        //create color map and normal map
        gl.glEnable(GL.GL_TEXTURE_2D);
        normalMap = createTextureMap("CS752_NormalMap.jpg");
        colorMap = createTextureMap("CS752_ColorMap.jpg");
        
        //enable light 
        gl.glEnable(GL.GL_LIGHTING);
        gl.glEnable(GL.GL_LIGHT0);
       
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, gray, 0);
        gl.glMaterialfv(GL.GL_FRONT, GL.GL_DIFFUSE, gray, 0);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, gray, 0);
        gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT, gray, 0);
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, white, 0);
        gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, white, 0);
        gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, shininess, 0);
        
        gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);
        
        
        //create shader program
        try 
        {
            attachShaders(gl,"CS752_VS.txt","CS752_FS.txt");
        } 
        catch (Exception e) 
        {
            e.printStackTrace();
        }
        
        surface = new CS752_Surface();
        
    }
    
    private Texture createTextureMap(String filepath) 
    {
        Texture tex = null;
        InputStream input = CS752_Screen.class.getResourceAsStream(filepath);
        
        try{
            //load texture
            tex = TextureIO.newTexture(input, true, filepath.substring(filepath.indexOf(".")));
            tex.bind();
        }
        catch(IOException ex)
        {
            System.err.println(ex);
        }
        
        return tex;
    }

    public void display(GLAutoDrawable drawable) 
    {
        
        GL gl = drawable.getGL();
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        
        //apply normal map
        gl.glActiveTexture(GL.GL_TEXTURE0);
        gl.glUniform1i(gl.glGetUniformLocation(shaderprogram, "normalMap"),0);
        gl.glBindTexture(GL.GL_TEXTURE_2D, normalMap.getTextureObject());
        
        //apply color map
        gl.glActiveTexture(GL.GL_TEXTURE1);
        gl.glUniform1i(gl.glGetUniformLocation(shaderprogram, "colorMap"),1);
        gl.glBindTexture(GL.GL_TEXTURE_2D, colorMap.getTextureObject());
        
        //draw surface
        gl.glLoadIdentity();
        
        gl.glPushMatrix();
        
        gl.glRotatef(fixedRotationX + rotationX, 1.0f, 0.0f, 0.0f);
        gl.glRotatef(fixedRotationY + rotationY, 0.0f, 1.0f, 0.0f);
        gl.glRotatef(fixedRotationZ + rotationZ, 0.0f, 1.0f, 0.0f);
        gl.glScalef(6, 6, 6);
        gl.glTranslatef(-0.5f, -0.5f, 0.0f);

        surface.draw(gl, shaderprogram);
        
        gl.glPopMatrix();
        gl.glFlush();
        
    }

    public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height) 
    {
        GL gl = drawable.getGL();
        gl.glViewport(0, 0, width, height);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        if (width <= (height * 2))
        {
            gl.glOrtho(-6.0, 6.0, -3.0 * ((float) height * 2) / (float) width, 3.0 * ((float) height * 2) / (float) width, -10.0, 10.0);
        }
        else
        {
            gl.glOrtho(-6.0 * (float) width / ((float) height * 2), 6.0 * (float) width / ((float) height * 2), -3.0, 3.0, -10.0, 10.0);
        }
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();
        gl.glOrtho(0, width, 0, height, -1.0, 1.0);
    }

    public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {}
    

    private void attachShaders(GL gl, String vertexFilename, String fragmentFilename) throws Exception 
    {
        //attach vertex shader
        String[] vertexSource = new String[] {readFile(vertexFilename)};
        int vertexShaderProgram = gl.glCreateShader(GL.GL_VERTEX_SHADER);
        gl.glShaderSource(vertexShaderProgram, 1, vertexSource, null, 0);
        gl.glCompileShader(vertexShaderProgram);

        //attach fragement shader
        String[] fragmentSource = new String[] {readFile(fragmentFilename)};
        int fragmentShaderProgram = gl.glCreateShader(GL.GL_FRAGMENT_SHADER);
        gl.glShaderSource(fragmentShaderProgram, 1, fragmentSource, null, 0);
        gl.glCompileShader(fragmentShaderProgram);

        //link shaders
        shaderprogram = gl.glCreateProgram();
        gl.glAttachShader(shaderprogram, vertexShaderProgram);
        gl.glAttachShader(shaderprogram, fragmentShaderProgram);
        gl.glLinkProgram(shaderprogram);
        gl.glValidateProgram(shaderprogram);
        IntBuffer intBuffer = IntBuffer.allocate(1);
        gl.glGetProgramiv(shaderprogram, GL.GL_LINK_STATUS,intBuffer);
        if (intBuffer.get(0)!=1)
        {
            gl.glGetProgramiv(shaderprogram, GL.GL_INFO_LOG_LENGTH,intBuffer);
            int size = intBuffer.get(0);
            System.err.println("Program link error: ");
            if (size>0)
            {
                ByteBuffer byteBuffer = ByteBuffer.allocate(size);
                gl.glGetProgramInfoLog(shaderprogram, size, intBuffer, byteBuffer);
                for (byte b:byteBuffer.array())
                {
                    System.err.print((char)b);
                }
            } 
            else 
            {
                System.out.println("Unknown");
            }
            System.exit(1);
        }
        
        //enable shader program
        gl.glUseProgram(shaderprogram);
    }

    private String readFile(String filename)
    {
        String outputString = "";
        InputStream input = CS752_Screen.class.getResourceAsStream(filename);
        
        Scanner scanner = new Scanner(input);
        outputString = scanner.useDelimiter("\\Z").next();
        
        return outputString;
    }

    public void mouseClicked(MouseEvent me) {}

    public void mousePressed(MouseEvent me) 
    {
        // Left click = select active control point
        // Right click = rotate camera
        if(me.getButton() == MouseEvent.BUTTON1)
            cmd = SELECT;
        else
            cmd = NOTHING;
        
        mouseX = me.getX();
        mouseY = me.getY();
    }

    public void mouseReleased(MouseEvent me) 
    {
        //release rotation
        if(cmd == NOTHING)
        {
            rotationY = 0.0f;
            rotationX = 0.0f;
            rotationZ = 0.0f;
            
            //lock in rotation
            if(me.isShiftDown())
            {
                fixedRotationX += me.getY() - mouseY;
                fixedRotationZ += me.getX() - mouseX;
            }
            else
            {
                fixedRotationX += me.getY() - mouseY;
                fixedRotationY += me.getX() - mouseX;
            }
        }
    }
    public void mouseEntered(MouseEvent me) {}
    public void mouseExited(MouseEvent me) {}
    public void mouseMoved(MouseEvent me) {}
    public void mouseDragged(MouseEvent me) 
    {
        // drag camera
        if(cmd == NOTHING)
        {
            if(me.isShiftDown())
            {
                rotationZ = me.getX() - mouseX;
                rotationX = me.getY() - mouseY;
            }
            else
            {
                rotationY = me.getX() - mouseX;
                rotationX = me.getY() - mouseY;
            }
        }
        
        // drag control point
        if(cmd == SELECT)
        {
            if(me.isShiftDown())
            {
                surface.movePoint(0, -(me.getX() - mouseX) * 0.001f, 0);
                surface.movePoint(-(me.getY() - mouseY) * 0.001f, 0, 0);
                mouseX = me.getX();
                mouseY = me.getY();
            }
            else
            {
                surface.movePoint(0, 0, -(me.getX() - mouseX) * 0.001f);
                surface.movePoint(0, -(me.getY() - mouseY) * 0.001f, 0);
                mouseX = me.getX();
                mouseY = me.getY();
            }
            
            //recalculate verticies when control point is moved
            surface.calculateVerticies();
        }
    }
    

    public void keyTyped(KeyEvent ke) 
    {
        // select next control point when space bar is typed
        if(ke.getKeyChar() == ' ')
            surface.selectNextPoint();
    }

    public void keyPressed(KeyEvent ke) 
    {
    }

    public void keyReleased(KeyEvent ke) 
    {
    }
    
}
